/*
*	bc module. Bytecode manipulation routines
*
*	History:
*		2001/02/23 - ndc - created
*/

#if !defined(_bc_h)
#define	_bc_h

#include "sys.h"
#include "pmem.h"
#include "kw.h"

#define	BC_ALLOC_INCR			1024
#define	BC_MAX_STORE_SIZE		0x7FFF

typedef	struct {
	MemHandle	mem_h;		// handle
	byte		*ptr;		// pointer
	word		cp;			// current position (used by readers not writers)
	word		size;		// allocation size
	word		count;		// current size		(used by writers as the current position)
	}	bc_t;

void	sc_raise(const char *fmt, ...) 			SEC(BCSCAN);

void	bc_create(bc_t *bc)						SEC(BCSCAN);
void	bc_destroy(bc_t *bc)  					SEC(BCSCAN);
void	bc_resize(bc_t *bc, dword newsize)		SEC(BCSCAN);

void	bc_add1 (bc_t *bc, byte code) 			SEC(BCSCAN);
void	bc_add2c(bc_t *bc, byte code, byte v)	SEC(BCSCAN);
#define	bc_add2(a,b,c)	bc_add2c((a),(b),(c))
void	bc_add1i(bc_t *bc, word p1)				SEC(BCSCAN);
void	bc_add2i(bc_t *bc, byte code, word p1)	SEC(BCSCAN);
void	bc_add2l(bc_t *bc, byte code, long p1)	SEC(BCSCAN);
void	bc_add2d(bc_t *bc, byte code, double p1) SEC(BCSCAN);
void	bc_add2s(bc_t *bc, byte code, const char *p1) SEC(BCSCAN);
char*	bc_store_string(bc_t *bc, char *src)	SEC(BCSCAN);
void	bc_eoc(bc_t *bc)						SEC(BCSCAN);

void	bc_append(bc_t *dest, bc_t *src)		SEC(BCSCAN);
void	bc_add_n(bc_t *dest, byte *src, word n)		SEC(BCSCAN);

#endif

