/*
*	SmallBASIC IDE
*
*	2000-05-26, Nicholas Christopoulos
*
*	This program is distributed under the terms of the GPL v2.0 or later
*	Download the GNU Public License (GPL) from www.gnu.org
*/

/*
*	Each DOC is a basic module
*	Each module has 
*		record 0 - type: info_t, the header
*		record 1 - type: sec_t,  the main code section
*		record n - type: sec_t,  code section
*
* 	Each section can be a function or a procedure.
* 	The maximum size of each section is 32KB (FIELD limit).
*/

#include "sys.h"
#include "sbpad.h"
#include "smbas.h"
#include "panic.h"
#include "str.h"
#include "ui.h"
#include "device.h"

#define	APPL	0x6170706C
#define	SmBa	0x536D4261

#define	idBASC	0x42415343
#define	idTEXT	0x54455854
#define	idDATA	0x44415441
#define idUFST  0x55465354

#define	idLNCH	0x6C6E6368

#define	MAXTEXTSIZE		0x7fff		// 32K
#define	MAXSTR			256

typedef char*		char_p;
typedef DmOpenRef	file_t;
#define	fatal(c,m)	ErrNonFatalDisplayIf((c), (m))

void savePrefs() SEC(IDE);

////////////////
extern byte os_charset;
//extern word os_cclabs1;
//extern word os_ccpass2;
////////////////

DWord	OldScreenDepth;

//#ifndef SysAppLaunchCmdOpenDBType
//typedef struct {
//	word	cardNo;
//	LocalID dbID;
//	} SysAppLaunchCmdOpenDBType;
//
//#define sysAppLaunchCmdOpenDB	52
//#endif

/*
*	sbpad prefs
*/
typedef struct {
	char	lastFileName[64];
	word	lastCat;
	word	lastDocSec;
	word	lastDocPos;
	word	lastDocFont;
	word	lastDocForm;
	word	font;
	word	charset;
	word	flags;
	word	cclabs1;
	word	ccpass2;
	char	reserver[64-20];
	} sbpref_t;

/*
*	doc header 
*/
typedef struct {
		byte	sign;				// always 'H'
		byte	version;			// 
		byte	unsued;
		byte	category;			//
		} info_t;

/*
*	doc section
*/
typedef struct {
	byte	sign;					// always = 'S'
	byte	unused;					//
	word	version;				// 
	word	flags;					// 1 for main section
	char	name[64];
	} sec_t;

//
static sbpref_t		sbpref;

/* current doc (for editor) */
static file_t		curFile;
static info_t		curFileHeader;
static char			curFileName[64];			// current filename (source file)
static char			curFSVName[64];				// current filename (file viewer)
static int			curSection;
static byte			modified = 0;

/* current doc section list */
static char_p		*secListStr;
static sec_t		*secList;
static int			secCount;

/* doc list variable */
static char_p *docTable;
static int	docCount;

/* user-file list  */
static char_p *ufstTable = NULL;
static int	ufstCount = 0;

static char	lastSearchString[MAXSTR];		// last search string
static int  autoReturnToForm;	  			// 

static char	temp[MAXSTR];

/// globals of ViewForm
static int	lastDocFont;
static int	lastDocPageSize;
static word	lastDocPos;
static word lastDocSec;
static word lastFSV;
static char	smtitle[64];

/// globals for categories
static char_p	*catList;
static int		catCount;
static dword	catUnfiledID;
static dword	curCat;
static MemHandle gHK40FontHandle;
static MemHandle gSM10FontHandle;
static MemHandle gDOSFontHandle;
static MemHandle gDOS4FontHandle;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int	BasicInputBox(char *title, char *dest, int size)
{
	FormPtr		frm, prv_frm;
	FieldPtr  	text_box;
	FormActiveStateType	stateP;
	VoidHand	text_h;
	char		*text_p;
	int			r, l;

	*dest = '\0';
	prv_frm = FrmGetActiveForm();
	FrmSaveActiveState(&stateP);


	frm = FrmInitForm(InputBoxForm);
	text_box = (FieldPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fldInputBox));
	FrmSetActiveForm(frm);
	FrmSetTitle(frm, title);
	FrmSetFocus(frm, FrmGetObjectIndex(frm, fldInputBox));
	FldSetInsPtPosition(text_box, 0);
	InsPtEnable(true);
	FldGrabFocus(text_box);

	// set value	
	if	( (r = FrmDoDialog(frm)) == btnOK )	{

		text_h = (VoidHand) FldGetTextHandle(text_box);
		if	( text_h )	{
			text_p = (char *) MemHandleLock(text_h);
												   
			if	( text_p )	{
				l = strlen(text_p);

				if	( l )	{
					if	( l >= size )	{
						StrNCopy(dest, text_p, size-1);
						dest[size] = '\0';
						}
					else
						StrCopy(dest, text_p);
					}

				MemHandleUnlock(text_h);
				}
			}
		}

	if ( prv_frm)	{
		FrmRestoreActiveState(&stateP);
		FrmSetActiveForm(prv_frm);
		}

	FrmDeleteForm(frm);
	return ( r == btnOK );
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////// BEAM
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Err	beamSendData(const void *data, dword *size, void *userdata)
{
	Err		err;

	*size = ExgSend((ExgSocketPtr) userdata, (void *) data, *size, &err);
	return err;
}

void beamSendDB(const char *file)
{
	char	*name;
	ExgSocketType	exsock;
	Err		err;
	LocalID	lid;

	lid = DmFindDatabase(0, (char *) file);
	if	( !lid )
		return;

	name = tmp_alloc(128);
	strcpy(name, file);
    DmDatabaseInfo(0, lid, name, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

	MemSet(&exsock, sizeof(exsock), 0);
	exsock.description = name;
	exsock.name = name;
	exsock.target = SmBa;
	err = ExgPut(&exsock);
	if	( !err )	{
		err = ExgDBWrite(beamSendData, &exsock, NULL, lid, 0);
		err = ExgDisconnect(&exsock, err);
		}
	tmp_free(name);
}

/*
*	BEAM: get database
*/

Err		readthunk(void *datap, unsigned long *sizep, void *rock)
{
	Err 			err = 0;
	ExgSocketType	*sock = rock;
	dword			count;

	count = *sizep;
	count = ExgReceive(sock, datap, count, &err);
	*sizep = count;
	return err;
}

Boolean	deletethunk(const char *name, Word version, word card, LocalID lid, void *rock)
{
	Err		err;
	unsigned short attr;

	err = DmDatabaseInfo(card, lid, NULL, &attr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	if ( !err )	{
		DmDeleteDatabase(0, lid);
		return 1;
		}
	return 0;
}

Err		beamGetDB(ExgSocketPtr sock, byte gavail)
{
	Err		err = 0;
	LocalID lid;
	unsigned short card = 0;
	byte	wantreset = 0;

	err = ExgDBRead(readthunk, deletethunk, sock, &lid, card, &wantreset, 1);
	err = ExgDisconnect(sock, err);

	sock->goToCreator = 0;

	if ( !err )	{
		if ( wantreset ) {
		    int res = FrmAlert(AlertWantReset);
		    if (res == 1)
		      SysReset();
			}
		}

	if ( gavail )	{
//		refresh doc list
		if	( sbpref.lastDocForm == MainForm )	{
			EventType	ev;

			ev.eType = evtRefreshDOCs;
			EvtAddEventToQueue(&ev);
			}
		} 
	return err;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int		dbExist(const char *file) SEC(IDE);
int		dbExist(const char *file)
{
	return (DmFindDatabase(0, (char*) file) != 0);
}

/*
*/
file_t	dbOpen(const char *fileName) SEC(IDE);
file_t	dbOpen(const char *fileName)
{
	LocalID		lid;
	file_t		ref;
	
	// open
	lid = DmFindDatabase(0, (char *) fileName);
	ref = DmOpenDatabase(0, lid, dmModeReadWrite);
	fatal(ref == 0, "dbOpen(): can't open file");
	return ref;
}

/*
*/
file_t	dbOpenAlw(const char *fileName, dword creator, dword type) SEC(IDE);
file_t	dbOpenAlw(const char *fileName, dword creator, dword type)
{
	LocalID		lid;
	file_t		ref;
	int			d;
	
	// open
	lid = DmFindDatabase(0, (char *) fileName);
	if	( !lid )	{ // create
		d = DmCreateDatabase(0, fileName, creator, type, false);
		fatal(d != 0, "dbOpwnAlw(): can't create file");
		lid = DmFindDatabase(0, (char *) fileName);
		}

	ref = DmOpenDatabase(0, lid, dmModeReadWrite);
	fatal(ref == 0, "dbOpenAlw(): can't open file");
	return ref;
}

/*
*/
void	dbClose(file_t f) SEC(IDE);
void	dbClose(file_t f)
{
	DmCloseDatabase(f);
}

/*
*	writes a new record at the end of the file
*/
word	dbWrite(file_t f, word record, void *new_rec, word size) SEC(IDE);
word	dbWrite(file_t f, word record, void *new_rec, word size)
{
	word		recIndex;
	mem_t		rec_h;
	void		*pre_rec;
	int		src_sz;
	int		d;

	if	( record >= DmNumRecords(f) )		{	// new record
		recIndex = dmMaxRecordIndex;
		rec_h = DmNewRecord(f, &recIndex, size);
		fatal(rec_h == 0, "dbWrite(): Can't create record");
		}
	else	{	// Update record
		recIndex = record;
		rec_h = DmGetRecord(f, recIndex);
		fatal(rec_h == 0, "dbWrite(): Can't open record");
		src_sz = MemHandleSize(rec_h);
		if	( src_sz != size )
			MemHandleResize(rec_h, size);
		}

	pre_rec = mem_lock(rec_h);
	fatal( pre_rec == 0,        "dbWrite(): Can't lock record!");
	d = DmWrite(pre_rec, 0, new_rec, size);
	fatal(d != 0, "dbWrite(): DmWrite return error.");
	mem_unlock(rec_h);
	d = DmReleaseRecord(f, recIndex, 1);
	fatal(d != 0, "dbWrite(): Can't release record.");
	return recIndex;
}

/*
*	writes a new record at the end of the file
*/
word	dbWriteSeg(file_t f, word record, word offset, void *ptr, word size) SEC(IDE);
word	dbWriteSeg(file_t f, word record, word offset, void *ptr, word size)
{
	word		recIndex;
	mem_t		rec_h;
	void		*pre_rec;
	int		src_sz;
	int		d;

	if	( record >= DmNumRecords(f) )		{	// new record
		recIndex = dmMaxRecordIndex;
		rec_h = DmNewRecord(f, &recIndex, size);
		fatal(rec_h == 0, "dbWriteSeg(): Can't create record");
		}
	else	{	// Update record
		recIndex = record;
		rec_h = DmGetRecord(f, recIndex);
		fatal(rec_h == 0, "dbWriteSeg(): Can't open record");
		src_sz = MemHandleSize(rec_h);
		if	( src_sz < (size+offset) )
			MemHandleResize(rec_h, (size+offset));
		}

	pre_rec = mem_lock(rec_h);
	fatal( pre_rec, "dbWriteSeg(): Can't lock record!");
	d = DmWrite(pre_rec, offset, ptr, size);
	fatal( d != 0, "dbWriteSeg(): DmWrite return error.");
	mem_unlock(rec_h);
	d = DmReleaseRecord(f, recIndex, 1);
	fatal(d != 0, "dbWriteSeg(): Can't release record.");
	return recIndex;
}

/*
*/
word	dbRecSize(file_t f, word record) SEC(IDE);
word	dbRecSize(file_t f, word record)
{
	VoidPtr		rec_p = NULL;
	VoidHand	rec_h = NULL;
	word		len;

	rec_h = DmGetRecord(f, record);
	fatal(d == 0, "dbRecSize(): can't get record");
	rec_p = mem_lock(rec_h);
	fatal(rec_p == NULL, "dbRecSize(): can't lock record");

	len = MemHandleSize(rec_h);
	mem_unlock(rec_h);
	DmReleaseRecord(f, record, 0);
	return len;
}

/*
*/
word	dbRead(file_t f, word record, void *buff, word size) SEC(IDE);
word	dbRead(file_t f, word record, void *buff, word size)
{
	VoidPtr		rec_p = NULL;
	VoidHand	rec_h = NULL;
	word		len;

	rec_h = DmGetRecord(f, record);
	fatal(rec_h == 0, "dbRead(): can't get record");
	rec_p = mem_lock(rec_h);
	fatal(rec_p == NULL, "dbRead(): can't lock record");

	len = MemHandleSize(rec_h);
	if	( len > size )
		len = size;

	MemMove(buff, rec_p, len);

	mem_unlock(rec_h);
	DmReleaseRecord(f, record, 0);
	return len;
}

/*
*/
word	dbReadSeg(file_t f, word record, void *buff, word offset, word size) SEC(IDE);
word	dbReadSeg(file_t f, word record, void *buff, word offset, word size)
{
	VoidPtr		rec_p = NULL;
	VoidHand	rec_h = NULL;
	word		len;

	rec_h = DmGetRecord(f, record);
	fatal(rec_h == 0, "dbReadSeg(): can't get record");
	rec_p = mem_lock(rec_h);
	fatal(rec_p == NULL, "dbReadSeg(): can't lock record");

	len = MemHandleSize(rec_h);
	if	( len > size )
		len = size;

	MemMove(buff, rec_p+offset, len);

	mem_unlock(rec_h);
	DmReleaseRecord(f, record, 0);
	return len;
}

/*
*/
int		dbRemove(const char *fileName) SEC(IDE);
int		dbRemove(const char *fileName)
{
	LocalID		lid;

	lid = DmFindDatabase(0, (char *) fileName);
	if	( !lid )
		return 0;
	return DmDeleteDatabase(0, lid);
}

/*
*/
word	dbRecCount(file_t f) SEC(IDE);
word	dbRecCount(file_t f)
{
	return DmNumRecords(f);
}

/*
*/
int		dbCopy(const char *oldName, const char *newName) SEC(IDE);
int		dbCopy(const char *oldName, const char *newName)
{
	file_t		fin, fout;
	word		recs, i, len;
	char		*ptr;

	fin = dbOpen(oldName);
	if	( dbExist(newName) ) dbRemove(newName);
	fout = dbOpenAlw(newName, SmBa /* idBASC */, idTEXT);

	// copy
	recs = dbRecCount(fin);
	for ( i = 0; i < recs; i ++ )	{
		len = dbRecSize(fin, i);
		ptr = tmp_alloc(len);
		dbRead(fin, i, ptr, len);

		dbWrite(fout, 0xFFFF, ptr, len);
		tmp_free(ptr);
		}

	// clean up
	dbClose(fin);
	dbClose(fout);
	return 1;
}

/*
*/
void	dbRename(const char *oldName, const char *newName) SEC(IDE);
void	dbRename(const char *oldName, const char *newName)
{
	if	( dbCopy(oldName, newName) )
		dbRemove(oldName);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void	mkSmallTitle(char *src, char *dst, int size, int font) SEC(IDE);
void	mkSmallTitle(char *src, char *dst, int size, int font)
{
	int		l, x, f;

	f = FntGetFont();
	FntSetFont(font);
	x = FntWordWrap(src, size);
	FntSetFont(f);

	l = StrLen(src);
	if	( l > x )	{
		MemMove(dst, src, x);
		dst[x] = '\x85';
		dst[x+1] = '\0';
		}
	else
		StrCopy(dst, src);
}

/*
*/
static void buildSecList() SEC(IDE);
static void buildSecList()
{
	byte	   	*ptr;
	int			recs, i, len;

	ErrNonFatalDisplayIf(!curFile, "BuildSecList(): database is not opened");

	recs = dbRecCount(curFile);

	secList = tmp_alloc(sizeof(sec_t) * recs);
	secListStr = tmp_alloc(sizeof(char_p) * recs);

	for ( i = 0; i < recs; i ++ )	{
		len = dbRecSize(curFile, i);
		ptr = tmp_alloc(len);
		dbRead(curFile, i, ptr, len);

		if	( *ptr == 'S' )	{
			MemMove(&secList[secCount], ptr, sizeof(sec_t));
			secListStr[secCount] = secList[secCount].name;
			secCount ++;
			}

		tmp_free(ptr);
		}
}

/*
*/
void destroySecList() SEC(IDE);
void destroySecList()
{
	if	( secList )	{
		tmp_free(secListStr);
		tmp_free(secList);
		secListStr = NULL;
		secList = NULL;
		secCount = 0;
		}
}

/*
*	destroy the list of user-filenames
*/
//static void destroyUserFileList() SEC(IDE);
void	destroyUserFileList()
{
	int		i;

	if	( ufstTable )	{
		for ( i = 0; i < ufstCount; i ++ )
			MemPtrFree(ufstTable[i]);
		tmp_free(ufstTable);
		ufstTable = NULL;
		ufstCount = 0;
		}
}

/*
*/
static void buildUserFileList() SEC(IDE);
static void buildUserFileList()
{
	int			dbCount, i;
	dword		dbType, dbCreator;
	LocalID		LID;

	destroyUserFileList();

	dbCount = DmNumDatabases(0);
	ufstTable = tmp_alloc(dbCount * sizeof(char_p));
	ufstCount = 0;
	for ( i = 0; i < dbCount; i++ )	{
		LID = DmGetDatabase(0, i);
		
		if	( LID )	{
			temp[0] = '\0';
			if	( DmDatabaseInfo(0, LID,
				temp,
				NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
				&dbType, &dbCreator ) == 0 )	{

				if ( dbCreator == SmBa && dbType == idUFST ) {
					ufstTable[ufstCount] = (char *) MemPtrNew(StrLen(temp)+1);
					ErrNonFatalDisplayIf(!ufstTable[ufstCount], "out of memory");
					StrCopy(ufstTable[ufstCount], temp);
					ufstCount ++;
					} // dbType
				} // DmDatabaseInfo
			} // LID
		} // for

	if	( ufstCount == 0 )	{
		ufstTable[ufstCount] = (char *) MemPtrNew(33);
		ErrNonFatalDisplayIf(!ufstTable[ufstCount], "out of memory");
		StrCopy(ufstTable[ufstCount], "NewFile");
		ufstCount ++;
		}
}

/*
*/
mem_t	loadSection(word secIndex) SEC(IDE);
mem_t	loadSection(word secIndex)
{
	word		len, rec_n;
	sec_t		*sec_p;
	mem_t		h;
	char		*ptr;

	fatal(!curFile, "loadSection(): database is not opened");
	fatal(secIndex > secCount, "loadSection(): page out of range");

	rec_n = secIndex + 1;
	sec_p = &secList[secIndex];
	len = dbRecSize(curFile, rec_n);

	h = mem_alloc(len - sizeof(sec_t));
	ptr = mem_lock(h);
	dbReadSeg(curFile, rec_n, ptr, sizeof(sec_t), len - sizeof(sec_t));
	mem_unlock(h);
	return h;
}

/*
*/
mem_t	loadFSV(const char *name) SEC(IDE);
mem_t	loadFSV(const char *name)
{
	mem_t		h = NULL;
	char		*ptr;
	dword		len;
	FileHand	f;
	Err			ferr;

	f = FileOpen(0, (char *)name, idUFST, SmBa, fileModeReadOnly, &ferr);
	if	( ferr == 0 )	{
		FileTell(f, &len, &ferr);
		if	( len > 32767 )	{
			len = 32767;
			warning("File size > 32KB");
			}
		h = mem_alloc(len+1);
		ptr = mem_lock(h);
		FileRead(f, ptr, len, 1, &ferr);
		ptr[len] = '\0';
		FileClose(f);
		mem_unlock(h);
		}
	return h;
}

/*
*/
int		createNewFSV(const char *name) SEC(IDE);
int		createNewFSV(const char *name)
{
	FileHand	f;
	Err			ferr;

	f = FileOpen(0, (char *)name, idUFST, SmBa, fileModeAppend, &ferr);
	if	( ferr == 0 )	
		FileClose(f);
	return (ferr != 0);
}

/*
*/
int		saveFSV(const char *name) SEC(IDE);
int		saveFSV(const char *name)
{
	char		*text;
	FileHand	f;
	Err			ferr;

	FileDelete(0, (char *)name);
	f = FileOpen(0, (char *)name, idUFST, SmBa, fileModeAppend, &ferr);
	if	( ferr == 0 )	{
		text = FldGetTextPtr(fld_ptr(fldTEXT));
		FileWrite(curFile, text, StrLen(text), 1, &ferr);
		FileClose(f);
		}
	return (ferr != 0);
}

/*
*/
void	openDOC(const char *fileName) SEC(IDE);
void	openDOC(const char *fileName)
{
	if	( curFile )
		dbClose(curFile);

	modified = 0;

	curFile = dbOpen(fileName);

	// header
	dbRead(curFile, 0, &curFileHeader, sizeof(curFileHeader));

	// build the section list
	destroySecList();
	buildSecList();
}

/*
*/
void	closeDOC() SEC(IDE);
void	closeDOC()
{
	if	( curFile )	{
		word	attr;
	
		destroySecList();
		dbClose(curFile);

		if	( modified )	{
			LocalID		lid;

			lid = DmFindDatabase(0, (char *) curFileName);
			if	( lid )	{
				DmDatabaseInfo(0, lid, NULL, &attr, NULL, NULL,  NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				attr |= dmHdrAttrBackup;
				DmSetDatabaseInfo(0, lid, NULL, &attr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				modified = 0;
				}
			}

		curFile = 0;
		}
}

/*
*/
void	closeFSV() SEC(IDE);
void	closeFSV()
{
	destroyUserFileList();
}

/*
*	Creates an empty doc
*/
int		createNewDOC(const char *fileName) SEC(IDE);
int		createNewDOC(const char *fileName)
{
	file_t		f;
	sec_t		sec;
	char		*defText;
	ULong		dts;
	DateTimeType cur_date;
	word		rec_n;

	if	( dbExist(fileName) )	{
		if	( FrmCustomAlert(ConfAlertID, "Overwrite '", (char *)fileName, "' ?") == 0 )	
			dbRemove(fileName);
		else
			return 0;
		}

	// default text 
	defText = tmp_alloc(64);
	dts = TimGetSeconds();
	TimSecondsToDateTime(dts, &cur_date);
	StrPrintF(defText, "' %s\n' %02d/%02d/%04d\n", fileName, (int) cur_date.day, (int) cur_date.month, cur_date.year);

	// create & open
	f = dbOpenAlw(fileName, SmBa /* idBASC */, idTEXT);

	// write header
	curFileHeader.sign = 'H';
	curFileHeader.version = 3;
	curFileHeader.category = (curCat == catUnfiledID) ? 0 : curCat;
	dbWrite(f, 0xFFFF, &curFileHeader, sizeof(curFileHeader));

	// write section
	sec.sign = 'S';
	sec.version = 3;
	sec.flags = 1;
	StrCopy(sec.name, "Main");
	rec_n = dbWriteSeg(f, 0xFFFF, 0, &sec, sizeof(sec_t));
	dbWriteSeg(f, rec_n, sizeof(sec_t), defText, StrLen(defText)+1);

	// close
	dbClose(f);
	tmp_free(defText);
	return 1;
}

/*
*	returns the document category
*/
dword	fastGetDocCat(LocalID lid)
{
	file_t		f;
	info_t		hdr;
	
	// open
	f = DmOpenDatabase(0, lid, dmModeReadWrite);
	fatal(f == 0, "fastGetDocCat: can't open file");
	if	( f )	{
		dbRead(f, 0, &hdr, sizeof(info_t));
		dbClose(f);
		return hdr.category;
		}
	return 0;
}


/*
*/
void 	updateDocCat(const char *file, dword category) SEC(IDE);
void	updateDocCat(const char *file, dword category)
{
	file_t		f;
	info_t		hdr;

	f = dbOpen(file);
	if	( f )	{
		dbRead(f, 0, &hdr, sizeof(info_t));
		hdr.category = category;
		dbWrite(f, 0, &hdr, sizeof(info_t));
		dbClose(f);
		}
}

/*
*/
void	saveSection(word secIndex) SEC(IDE);
void	saveSection(word secIndex)
{
	char		*text;

	fatal(!curFile, "saveSection(): database is not opened");
	fatal(secIndex > secCount, "saveSection(): page out of range");

	text = FldGetTextPtr(fld_ptr(fldTEXT));
	dbWriteSeg(curFile, secIndex + 1, sizeof(sec_t), text, StrLen(text)+1);
	modified = 1;	// needed to set the backup field
}

/*
*/
void	saveSectionName(word secIndex, char *name) SEC(IDE);
void	saveSectionName(word secIndex, char *name)
{
	dbWriteSeg(curFile, secIndex + 1, sizeof(sec_t)-64, name, StrLen(name)+1);
	modified = 1;	// needed to set the backup field
}

/*
*/
void 	createNewSec(char *name) SEC(IDE);
void	createNewSec(char *name)
{
	word		rec_n;
	sec_t		sec;
	char		defText[64];

	fatal(!curFile, "createNewSection(): database is not opened");

	mkSmallTitle(name, defText, 55, stdFont);
	StrCopy(sec.name, defText);

	StrCopy(defText, "' ");
	StrCat(defText, name);

	// write section
	sec.sign = 'S';
	sec.flags = 0;

	rec_n = dbWriteSeg(curFile, 0xFFFF, 0, &sec, sizeof(sec_t));
	dbWriteSeg(curFile, rec_n, sizeof(sec_t), defText, StrLen(defText)+1);
}

/*
*/
void	deleteSec(word idx) SEC(IDE);
void	deleteSec(word idx)
{
	fatal(!curFile, "deleteSec(): database is not opened");
	if	( idx > 0 )	
		{ fatal( (DmRemoveRecord(curFile, idx+1) != 0), "deleteSec(): can't delete record"); }
	else
		FrmCustomAlert(InfoAlertID, "You cannot delete the main section", "", "");
}

void	checkSaveDOC() SEC(IDE);
void	checkSaveDOC()
{
	if	( curFile )	{
		if	( FldDirty(fld_ptr(fldTEXT)) )	{
			if	( FrmCustomAlert(ConfAlertID, "The text has been modified. Save now ?", "", "") == 0 )	{
				FldCompactText(fld_ptr(fldTEXT));
				saveSection(curSection);
				}
			}
		}
}

void	checkSaveFSV() SEC(IDE);
void	checkSaveFSV()
{
	if	( FldDirty(fld_ptr(fldTEXT)) )	{
		if	( FrmCustomAlert(ConfAlertID, "The text has been modified. Save now ?", "", "") == 0 )	{
			FldCompactText(fld_ptr(fldTEXT));
			saveFSV(curFSVName);
			}
		}
}

/*
*/
int		dispSection(word secIndex) SEC(IDE);
int		dispSection(word secIndex)
{
	mem_t	new_h;

	fatal(!curFile, "dispSection(): database is not opened");

	if ( secIndex < secCount )	{
		checkSaveDOC();

		new_h = loadSection(secIndex);
		curSection = secIndex;
		setFieldHandle(fldTEXT, new_h, 1);

		return 0;
		}

	return 1;
}

/*
*/
int		dispFSV(const char *name) SEC(IDE);
int		dispFSV(const char *name)
{
	mem_t	new_h;
	char	*ptr;

	if	( dbExist(name) )	{
		// checkSaveDOC();
		new_h = loadFSV(name);
//		StrCopy(curFSVName, name);
		}
	else	{
		new_h = mem_alloc(128);
		ptr = mem_lock(new_h);
		StrPrintF(ptr, "File '%s' not found!", name);
		mem_unlock(new_h);
		}
	setFieldHandle(fldTEXT, new_h, 1);

	return 0;
}

void	setTextPos(word pos) SEC(IDE);
void	setTextPos(word pos)
{
	FieldPtr	fp;

	fp = fld_ptr(fldTEXT);
	FldSetScrollPosition(fp, pos);
	FldSetInsPtPosition(fp, pos);
//	FldSetInsertionPoint(fp, pos);
	FldDrawField(fp);
	FldGrabFocus(fp);
}

word	getTextPos() SEC(IDE);
word	getTextPos()
{
	return FldGetInsPtPosition(fld_ptr(fldTEXT));
}

int 	copyDOC(const char *oldName, const char *newName) SEC(IDE);
int		copyDOC(const char *oldName, const char *newName)
{
	if	( !dbExist(oldName) )
		return 0;
    if	( !StrCompare(oldName,newName))	{
		FrmCustomAlert(InfoAlertID, "New name '", (char *) newName, "' same as old name.  No action taken.");
		return 0;
		}
	if	( dbExist(newName) )	{
		if	( FrmCustomAlert(ConfAlertID, "Overwrite '", (char *) newName, "' ?") != 0 )
			return 0;
		}

	return dbCopy(oldName, newName);
}

void	renameDOC(const char *oldName, const char *newName) SEC(IDE);
void	renameDOC(const char *oldName, const char *newName)
{
	if	( copyDOC(oldName, newName) )
		DmDeleteDatabase(0, DmFindDatabase(0, (char *) oldName));
}

// 0 = fail
int		Exec(const char *prog) SEC(IDE);
int		Exec(const char *prog)
{
	LocalID	lid;
	dword	progid;
	word	card;
	DmSearchStateType	state;

	progid  = ((dword) prog[0] << 24) + ((dword) prog[1] << 16) + ((dword) prog[2] << 8) + (dword) prog[3];
	if	( progid == idLNCH )	{
		EventType	ev;

		MemSet(&ev, sizeof(EventType), 0);
		ev.eType = keyDownEvent;
		ev.data.keyDown.modifiers = commandKeyMask;
		ev.data.keyDown.chr = vchrLaunch;
		EvtAddEventToQueue(&ev);
		return 1;
		}
	else	{
		if	( DmGetNextDatabaseByTypeCreator(true, &state, APPL, progid, true, &card, &lid) == 0 )
			return (SysUIAppSwitch(card, lid, sysAppLaunchCmdNormalLaunch, NULL) == 0);
		}
	return 0;
}

/*
*	destroy the list of 'basic-source' filenames
*/
//static void destroyDocList() SEC(IDE);
void	destroyDocList()
{
	int		i;

	if	( docTable )	{
		for ( i = 0; i < docCount; i ++ )
			MemPtrFree(docTable[i]);
		tmp_free(docTable);
		docTable = NULL;
		docCount = 0;
		}
}

/*
*	build the list of 'basic-source' filenames
*/
//static void buildDocList() SEC(IDE);
void	buildDocList()
{
	int			dbCount, i;
	dword		dbType, dbCat, dbCreator;
	LocalID		LID;

	destroyDocList();

	dbCount = DmNumDatabases(0);
	docTable = tmp_alloc(dbCount * sizeof(char_p));
	docCount = 0;
	for ( i = 0; i < dbCount; i++ )	{
		LID = DmGetDatabase(0, i);
		
		if	( LID )	{
			temp[0] = '\0';
			if	( DmDatabaseInfo(0, LID,
				temp,
				NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
				&dbType, &dbCreator ) == 0 )	{

				if ( (dbCreator == idBASC || dbCreator == SmBa) && dbType == idTEXT ) {
					dbCat = fastGetDocCat(LID);
					if	( curCat == 0 || dbCat == curCat || (dbCat == 0 && curCat == catUnfiledID) )	{
						docTable[docCount] = (char *) MemPtrNew(StrLen(temp)+1);
						ErrNonFatalDisplayIf(!docTable[docCount], "out of memory");
						StrCopy(docTable[docCount], temp);
						docCount ++;
						}
					} // dbType
				} // DmDatabaseInfo
			} // LID
		} // for
}

void	gotoLine(int line) SEC(IDE);
void	gotoLine(int line)
{
	char	*p, *text;
	int		c = 1, pos = 0, found = 0;

	if	( line > 0 )	{
		if	( line == 1 )
			setTextPos(0);

		text = FldGetTextPtr(fld_ptr(fldTEXT));
		p = text;

		while ( *p )	{
			if	( *p == '\n' )	{
				c ++;
				if	( c == line )	{
					setTextPos(pos+1);
					found ++;
					break;
					}
				}

			p ++;
			pos ++;
			}

		if	( !found )
			setTextPos(pos+1);
		}
}

void	searchFor(const char *src) SEC(IDE);
void	searchFor(const char *src)
{
	char	*p, *text;
	int		pos = 0;

	StrCopy(lastSearchString, src);
	text = FldGetTextPtr(fld_ptr(fldTEXT));
	p = text;

	while ( *p )	{
		if	( StrNCaselessCompare(p, lastSearchString, StrLen(lastSearchString)) == 0 )	{
			setTextPos(pos+1);
			break;
			}

		p ++;
		pos ++;
		}
}

void	searchAgain() SEC(IDE);
void	searchAgain()
{
	char	*p, *text;
	int		pos;

	text = FldGetTextPtr(fld_ptr(fldTEXT));
	p = text;
	p += ((pos = getTextPos()) + 1);

	while ( *p )	{
		if	( StrNCaselessCompare(p, lastSearchString, StrLen(lastSearchString)) == 0 )	{
			setTextPos(pos+1);
			break;
			}

		p ++;
		pos ++;
		}
}

void	mkPutFileExt(char *text)	SEC(IDE);
void	mkPutFileExt(char *text)
{
	char	*p;

	p = str_istr(text, ".bas");
	if	( p == NULL )	
		StrCat(text, ".bas");
}

//// EDIT /////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void	IDESaveStatus(const char *file, word lastSec, word lastPos, int font)	SEC(IDE);
void	IDESaveStatus(const char *file, word lastSec, word lastPos, int font)
{
	file_t	f;
	int		i, recs, found = 0;
	byte	*buff;

	f = dbOpenAlw("SmBa.IDE.DAT", SmBa, idDATA);
	buff = tmp_alloc(128);

	recs = DmNumRecords(f);
	for ( i = 0; i < recs; i ++ )	{
		dbRead(f, i, buff, 128);
		if	( StrCompare(buff, file) == 0 )	{
			buff[66] = (lastSec >> 8);		buff[67] = lastSec & 0xFF;
			buff[68] = (lastPos >> 8);		buff[69] = lastPos & 0xFF;
			buff[70] = (font >> 8);			buff[71] = font    & 0xFF;
			dbWrite(f, i, buff, 128);
			found = 1;
			break;
			}
		}

	if	( !found )	{
		StrCopy(buff, file);
		buff[64] = 0;		buff[65] = 1;
		buff[66] = (lastSec >> 8);		buff[67] = lastSec & 0xFF;
		buff[68] = (lastPos >> 8);		buff[69] = lastPos & 0xFF;
		buff[70] = (font >> 8);			buff[71] = font    & 0xFF;
		dbWrite(f, 0xFFFF, buff, 128);
		}

	tmp_free(buff);
	dbClose(f);
}

int		IDELoadStatus(const char *file, word *lastSec, word *lastPos, int *font) SEC(IDE);
int		IDELoadStatus(const char *file, word *lastSec, word *lastPos, int *font)
{
	file_t	f;
	int		i, recs, found = 0;
	byte	*buff;

	f = dbOpenAlw("SmBa.IDE.DAT", SmBa, idDATA);
	buff = tmp_alloc(128);

	recs = DmNumRecords(f);
	for ( i = 0; i < recs; i ++ )	{
		dbRead(f, i, buff, 128);
		if	( StrCompare(buff, file) == 0 )	{
			*lastSec = (buff[66] << 8) + buff[67];
			*lastPos = (buff[68] << 8) + buff[69];
			*font    = (buff[70] << 8) + buff[71];
			found = 1;
			break;
			}
		}

	tmp_free(buff);
	dbClose(f);
	return found;
}

Boolean		ViewFormHandleEventSEG(EventPtr e)	SEC(IDE);
Boolean		ViewFormHandleEventSEG(EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
	ListPtr	list;
    
    switch (e->eType) {
    case frmOpenEvent:
		frm = FrmGetActiveForm();
		mkSmallTitle(curFileName, smtitle, 60, boldFont);
		FrmSetTitle(frm, smtitle);
		FrmDrawForm(frm);

		sbpref.lastDocForm = ViewForm;

		openDOC(curFileName);

		if	( !IDELoadStatus(curFileName, &lastDocSec, &lastDocPos, &lastDocFont) )	{
			lastDocSec = 0;
			lastDocPos = 0;
			lastDocFont = sbpref.lastDocFont;
			}

		FldSetFont(GetFieldPtr(fldTEXT), lastDocFont);
		if	( lastDocFont == HK40CustomFont )	
			lastDocPageSize = 20;
		else if	( lastDocFont == SM10CustomFont )	
			lastDocPageSize = 13;
		else if	( lastDocFont == DOSCustomFont )	
			lastDocPageSize = 13;
		else if	( lastDocFont == DOS4CustomFont )	
			lastDocPageSize = 13;
		else if	( lastDocFont == 0 || lastDocFont == 1 )
			lastDocPageSize = 10;
		else if ( lastDocFont == 2 )
			lastDocPageSize = 8;
		else
			lastDocPageSize = 7;

		list = GetListPtr(lstSection);
		LstSetListChoices(list, secListStr, secCount);
		UpdatePopup(trgSection, lstSection, lastDocSec, secListStr);

		dispSection(lastDocSec);
		setTextPos(lastDocPos);

		autoReturnToForm = 0;

		FrmSetFocus(frm, FrmGetObjectIndex(frm, fldTEXT));

		handled = true;
		break;
	case popSelectEvent:
    	switch (e->data.popSelect.controlID) {
	    case trgSection:
			if	( lastDocSec != e->data.popSelect.selection )	{
				UpdatePopup(trgSection, lstSection, e->data.popSelect.selection, secListStr);
				lastDocSec = e->data.popSelect.selection;
				dispSection(lastDocSec);
				setTextPos(0);
				}
			handled = 1;
			break;
			}
		break;
    case menuEvent:
		switch(e->data.menu.itemID) {
		case mnuNew:
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);

			if	( BasicInputBox("New FileName:", temp, 255) )	{
				checkSaveDOC();

				mkPutFileExt(temp);
				if	( createNewDOC(temp) )	{
					openDOC(temp);
					
					lastDocSec = lastDocPos = 0;
					StrCopy(curFileName, temp);

					frm = FrmGetActiveForm();
					mkSmallTitle(curFileName, smtitle, 60, boldFont);
					FrmSetTitle(frm, smtitle);
					FrmDrawForm(frm);

					list = GetListPtr(lstSection);
					LstSetListChoices(list, secListStr, secCount);
					UpdatePopup(trgSection, lstSection, lastDocSec, secListStr);

					dispSection(lastDocSec);
					setTextPos(lastDocPos);
					}
				}
			break;
		case mnuBeam:
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
			checkSaveDOC();
			beamSendDB(curFileName);
			break;
		case mnuEdit:
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
			checkSaveDOC();

			closeDOC();
			FrmGotoForm(MainForm);
			break;
		case mnuRun:
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
			checkSaveDOC();

			autoReturnToForm = ViewForm;
			closeDOC();
			FrmGotoForm(SBOutForm);
			break;
		case mnuSave:
			if ( curFile )	{
				lastDocPos = getTextPos();
				IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
				saveSection(curSection);
				FldSetDirty(fld_ptr(fldTEXT), 0);
				}
			break;
		case mnuSaveAs:
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
			if ( curFile )	{
				if	( BasicInputBox("New FileName:", temp, 255) )	{
					frm = FrmGetActiveForm();
					closeDOC();

					mkPutFileExt(temp);
					copyDOC(curFileName, temp);

					StrCopy(curFileName, temp);
					openDOC(curFileName);

					list = GetListPtr(lstSection);
					LstSetListChoices(list, secListStr, secCount);
					UpdatePopup(trgSection, lstSection, lastDocSec, secListStr);

					mkSmallTitle(curFileName, smtitle, 60, boldFont);
					FrmSetTitle(frm, smtitle);
					FrmDrawForm(frm);

					saveSection(lastDocSec);
					setTextPos(lastDocPos);
					}
				}
			break;
		case mnuClose:
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
			checkSaveDOC();

			closeDOC();
			FrmGotoForm(MainForm);
			break;
		case mnuGoto:
			if	( BasicInputBox("Go to line...", temp, 64) )	{
				gotoLine(xstrtol(temp));
				lastDocPos = getTextPos();
				}
	      	break;
		case mnuSearch:
			if	( BasicInputBox("Search for:", temp, 255) )	{
				searchFor(temp);
				lastDocPos = getTextPos();
				}
	      	break;
		case mnuSearchAgain:
			searchAgain();
			lastDocPos = getTextPos();
			break;
		case mnuAbout:
			DoAboutForm();
	      	break;
		case mnuHelp:
			GPlmAlert(HelpStrTID, HelpStringID);
			break;

		case mnuHelpFS:
			GPlmAlert(HelpFSStrTID, HelpFSStringID);
			break;
		case mnuHelpMisc:
			GPlmAlert(HelpMiscStrTID, HelpMiscStringID);
			break;
		case mnuHelpSys:
			GPlmAlert(HelpSysStrTID, HelpSysStringID);
			break;


		case mnuHelpGraf:
			GPlmAlert(HelpGrafStrTID, HelpGrafStringID);
			break;
		case mnuHelpMS:
			GPlmAlert(HelpMSStrTID, HelpMSStringID);
			break;
		case mnuHelpCon:
			GPlmAlert(HelpConStrTID, HelpConStringID);
			break;
		case mnuUndo:
			FldUndo(fld_ptr(fldTEXT));
			break;
		case mnuCut:
			FldCut(fld_ptr(fldTEXT));
			break;
		case mnuCopy:
			FldCopy(fld_ptr(fldTEXT));
			break;
		case mnuSelAll:
			FldSetSelection(fld_ptr(fldTEXT), 0, 0xFFFF);
			break;
	   	case mnuPaste:			  
			FldPaste(fld_ptr(fldTEXT));
			break;
	   	case mnuKeyboard:
			SysKeyboardDialog(kbdDefault);
			break;
		 case mnuGraffiti:
			SysGraffitiReferenceDialog(referenceDefault);
			break;
	   	case mnuFont0:
			FldSetFont(fld_ptr(fldTEXT), (lastDocFont = stdFont));
			lastDocPageSize = 10;
			break;
	   	case mnuFont1:
			FldSetFont(fld_ptr(fldTEXT), (lastDocFont = boldFont));
			lastDocPageSize = 10;
			break;
	   	case mnuFont2:
			FldSetFont(fld_ptr(fldTEXT), (lastDocFont = largeFont));
			lastDocPageSize = 8;
			break;
	   	case mnuFont7:
			FldSetFont(fld_ptr(fldTEXT), (lastDocFont = largeBoldFont));
			lastDocPageSize = 7;
			break;
	   	case mnuFontHK:
			if	( os_charset == enc_utf8 )	{
				FldSetFont(fld_ptr(fldTEXT), (lastDocFont = HK40CustomFont));
				lastDocPageSize = 20;
				}
			else
				FrmAlert(AlertCFontMB);
			break;
	   	case mnuFontSM10:
			if	( os_charset == enc_utf8 )	{
				FldSetFont(fld_ptr(fldTEXT), (lastDocFont = SM10CustomFont));
				lastDocPageSize = 13;
				}
			else
				FrmAlert(AlertCFontMB);
			break;
	   	case mnuFontDOS:
			if	( os_charset == enc_utf8 )	{
				FldSetFont(fld_ptr(fldTEXT), (lastDocFont = DOSCustomFont));
				lastDocPageSize = 15;
				}
			else
				FrmAlert(AlertCFontMB);
			break;
	   	case mnuFontDOS4:
			if	( os_charset == enc_utf8 )	{
				FldSetFont(fld_ptr(fldTEXT), (lastDocFont = DOS4CustomFont));
				lastDocPageSize = 15;
				}
			else
				FrmAlert(AlertCFontMB);
			break;
		case	mnuNewSec:
			if	( BasicInputBox("Section Name:", temp, 255) )	{
				temp[32] = '\0';
				checkSaveDOC();

				createNewSec(temp);

				destroySecList();
				buildSecList();
				lastDocSec = secCount - 1;

				list = GetListPtr(lstSection);
				LstSetListChoices(list, secListStr, secCount);
				UpdatePopup(trgSection, lstSection, lastDocSec, secListStr);

				dispSection(lastDocSec);
				setTextPos(lastDocPos = 0);
				}
			break;
		case	mnuChangeSec:
			if	( BasicInputBox("New name:", temp, 255) )	{
				temp[32] = '\0';
				saveSectionName(curSection, temp);

				destroySecList();
				buildSecList();

				list = GetListPtr(lstSection);
				LstSetListChoices(list, secListStr, secCount);
				UpdatePopup(trgSection, lstSection, lastDocSec, secListStr);
				}
			break;
		case	mnuDelSec:
			if	( FrmCustomAlert(ConfAlertID, "Do you want to delete the current section ?", "", "") == 0 )		{
				dispSection(0);
				setTextPos((lastDocPos = 0));

				deleteSec(lastDocSec);

				destroySecList();
				buildSecList();

				lastDocSec = 0;
				list = GetListPtr(lstSection);
				LstSetListChoices(list, secListStr, secCount);
				UpdatePopup(trgSection, lstSection, lastDocSec, secListStr);
				}
			break;
		  	};
    	handled = true;
		break;

	case keyDownEvent:
		switch ( e->data.keyDown.chr )	{
		case	pageUpChr: /* Prev */
			if ( FldScrollable(fld_ptr(fldTEXT), winUp) ) 
				FldScrollField(fld_ptr(fldTEXT), lastDocPageSize, winUp);
			handled = true;
			break;
		case pageDownChr: /* next */
			if	( FldScrollable(fld_ptr(fldTEXT), winDown) )	
				FldScrollField(fld_ptr(fldTEXT), lastDocPageSize, winDown);
			handled = 1;
			break;
			};
		break;
    case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
		case	btnSFT:
			FldInsert(fld_ptr(fldTEXT), "\t", 1);
	    	handled = true;
			break;
		case	btnCLOSE:
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
			checkSaveDOC();

			closeDOC();
			FrmGotoForm(MainForm);
	    	handled = true;
			break;
		case	btnUP:
			if ( FldScrollable(fld_ptr(fldTEXT), winUp) ) 
				FldScrollField(fld_ptr(fldTEXT), 1, winUp);
	    	handled = true;
			break;
		case	btnDOWN:
			if	( FldScrollable(fld_ptr(fldTEXT), winDown) )	
				FldScrollField(fld_ptr(fldTEXT), 1, winDown);
	    	handled = true;
			break;
		case	btnPGUP:
			if ( FldScrollable(fld_ptr(fldTEXT), winUp) ) 
				FldScrollField(fld_ptr(fldTEXT), lastDocPageSize, winUp);
	    	handled = true;
			break;
		case	btnPGDN:
			if	( FldScrollable(fld_ptr(fldTEXT), winDown) )	
				FldScrollField(fld_ptr(fldTEXT), lastDocPageSize, winDown);
	    	handled = true;
			break;
		case	btnSAVE:
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
			saveSection(curSection);
			FldSetDirty(fld_ptr(fldTEXT), 0);
	    	handled = true;
			break;
		case btnQRUN:
			// SAVE
			lastDocPos = getTextPos();
			IDESaveStatus(curFileName, lastDocSec, lastDocPos, lastDocFont);
			saveSection(curSection);
			FldSetDirty(fld_ptr(fldTEXT), 0);

			// AND RUN
			autoReturnToForm = ViewForm;
			closeDOC();
			FrmGotoForm(SBOutForm);
			break;
			}
		break;		   

    default:
        break;
    	}

    return handled;
}

Boolean		ViewFormHandleEvent(EventPtr e)
{
	return ViewFormHandleEventSEG(e);
}

//// FILEVIEW /////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void	FSVSaveStatus(const char *file, word lastPos, int font)	SEC(IDE);
void	FSVSaveStatus(const char *file, word lastPos, int font)
{
	file_t	f;
	int		recs;
	byte	*buff;

	f = dbOpenAlw("SmBa.FSV.DAT", SmBa, idDATA);
	buff = tmp_alloc(128);

	recs = DmNumRecords(f);
	StrCopy(buff, file);
	buff[68] = (lastPos >> 8);		buff[69] = lastPos & 0xFF;
	buff[70] = (font >> 8);			buff[71] = font    & 0xFF;
	if	( recs > 0 )	
		dbWrite(f, 0, buff, 128);
	else
		dbWrite(f, 0xFFFF, buff, 128);

	tmp_free(buff);
	dbClose(f);
}

int		FSVLoadStatus(char *file, word *lastPos, int *font) SEC(IDE);
int		FSVLoadStatus(char *file, word *lastPos, int *font)
{
	file_t	f;
	int		recs, found = 0;
	byte	*buff;

	f = dbOpenAlw("SmBa.FSV.DAT", SmBa, idDATA);
	buff = tmp_alloc(128);

	recs = DmNumRecords(f);
	if	( recs > 0 )	{
		dbRead(f, 0, buff, 128);
		StrCopy(file, buff);
		*lastPos = (buff[68] << 8) + buff[69];
		*font    = (buff[70] << 8) + buff[71];
		found = 1;
		}
	else	{
		file[0] = '\0';
		*lastPos = 0;
		*font = 0;
		}

	tmp_free(buff);
	dbClose(f);
	return found;
}

Boolean		FSVFormHandleEventSEG(EventPtr e)	SEC(IDE);
Boolean		FSVFormHandleEventSEG(EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
	ListPtr	list;
	int		i;
    
    switch (e->eType) {
    case frmOpenEvent:
		frm = FrmGetActiveForm();
		FrmDrawForm(frm);

		buildUserFileList();

		if	( !FSVLoadStatus(curFSVName, &lastDocPos, &lastDocFont) )	{
			lastDocPos = 0;
			lastDocFont = 0;
			lastFSV = 0;
			}
		else	{
			lastFSV = 0;
			for ( i = 0; i < ufstCount; i ++ )	{
				if	( StrCompare(curFSVName, ufstTable[i]) == 0 )	{
					lastFSV = i;
					break;
					}
				}
			}

		FldSetFont(GetFieldPtr(fldTEXT), lastDocFont);
		if	( lastDocFont == HK40CustomFont )	
			lastDocPageSize = 20;
		else if	( lastDocFont == SM10CustomFont )	
			lastDocPageSize = 13;
		else if	( lastDocFont == DOSCustomFont )	
			lastDocPageSize = 13;
		else if	( lastDocFont == DOS4CustomFont )	
			lastDocPageSize = 13;
		else if	( lastDocFont == 0 || lastDocFont == 1 )
			lastDocPageSize = 10;
		else if ( lastDocFont == 2 )
			lastDocPageSize = 8;
		else
			lastDocPageSize = 7;

		list = GetListPtr(lstSection);
		LstSetListChoices(list, ufstTable, ufstCount);
		UpdatePopup(trgSection, lstSection, lastFSV, ufstTable);

		StrCopy(curFSVName, ufstTable[lastFSV]);
		dispFSV(curFSVName);
		setTextPos(lastDocPos);

		autoReturnToForm = 0;

		FrmSetFocus(frm, FrmGetObjectIndex(frm, fldTEXT));

		handled = true;
		break;
	case popSelectEvent:
    	switch (e->data.popSelect.controlID) {
	    case trgSection:
			if	( lastFSV != e->data.popSelect.selection )	{
				UpdatePopup(trgSection, lstSection, e->data.popSelect.selection, ufstTable);
				lastFSV = e->data.popSelect.selection;
				StrCopy(curFSVName, ufstTable[lastFSV]);
				dispFSV(curFSVName);
				setTextPos(0);
				}
			handled = 1;
			break;
			}
		break;
    case menuEvent:
		switch(e->data.menu.itemID) {
		case mnuNew:
			lastDocPos = getTextPos();
			FSVSaveStatus(curFSVName, lastDocPos, lastDocFont);

			if	( BasicInputBox("New FileName:", temp, 255) )	{
				checkSaveFSV();

				if	( createNewFSV(temp) )	{
					lastFSV = 0;
					for ( i = 0; i < ufstCount; i ++ )	{
						if	( StrCompare(curFSVName, ufstTable[i]) == 0 )	{
							lastFSV = i;
							break;
							}
						}

					list = GetListPtr(lstSection);
					LstSetListChoices(list, ufstTable, ufstCount);
					UpdatePopup(trgSection, lstSection, lastFSV, ufstTable);

					StrCopy(curFSVName, ufstTable[lastFSV]);
					dispFSV(curFSVName);
					setTextPos(lastDocPos);
					}
				}
			break;
		case mnuBeam:
			lastDocPos = getTextPos();
			FSVSaveStatus(curFSVName, lastDocPos, lastDocFont);
			checkSaveFSV();
			beamSendDB(curFSVName);
			break;
		case mnuEdit:
			lastDocPos = getTextPos();
			FSVSaveStatus(curFSVName, lastDocPos, lastDocFont);
			checkSaveFSV();

			closeFSV();
			FrmGotoForm(MainForm);
			break;
		case mnuSave:
			lastDocPos = getTextPos();
			FSVSaveStatus(curFSVName, lastDocPos, lastDocFont);
			saveFSV(curFSVName);
			FldSetDirty(fld_ptr(fldTEXT), 0);
			break;
		case mnuSaveAs:
			lastDocPos = getTextPos();
			FSVSaveStatus(curFSVName, lastDocPos, lastDocFont);
			if	( BasicInputBox("New FileName:", temp, 255) )	{
				frm = FrmGetActiveForm();
				saveFSV(temp);

				lastFSV = 0;
				for ( i = 0; i < ufstCount; i ++ )	{
					if	( StrCompare(curFSVName, ufstTable[i]) == 0 )	{
						lastFSV = i;
						break;
						}
					}

				list = GetListPtr(lstSection);
				LstSetListChoices(list, ufstTable, ufstCount);
				UpdatePopup(trgSection, lstSection, lastFSV, ufstTable);

				StrCopy(curFSVName, ufstTable[lastFSV]);
				dispFSV(curFSVName);
				setTextPos(lastDocPos);
				}
			break;
		case mnuClose:
			lastDocPos = getTextPos();
			FSVSaveStatus(curFSVName, lastDocPos, lastDocFont);
			checkSaveFSV();

			closeFSV();
			FrmGotoForm(MainForm);
			break;
		case mnuGoto:
			if	( BasicInputBox("Go to line...", temp, 64) )	{
				gotoLine(xstrtol(temp));
				lastDocPos = getTextPos();
				}
	      	break;
		case mnuSearch:
			if	( BasicInputBox("Search for:", temp, 255) )	{
				searchFor(temp);
				lastDocPos = getTextPos();
				}
	      	break;
		case mnuSearchAgain:
			searchAgain();
			lastDocPos = getTextPos();
			break;
		case mnuAbout:
			DoAboutForm();
	      	break;
		case mnuFileInfo:
	      	break;
		case mnuFSVHelp:
//			GPlmAlert(HelpFSVStrTID, HelpFSVStringID);
	      	break;
		case mnuUndo:
			FldUndo(fld_ptr(fldTEXT));
			break;
		case mnuCut:
			FldCut(fld_ptr(fldTEXT));
			break;
		case mnuCopy:
			FldCopy(fld_ptr(fldTEXT));
			break;
		case mnuSelAll:
			FldSetSelection(fld_ptr(fldTEXT), 0, 0xFFFF);
			break;
	   	case mnuPaste:			  
			FldPaste(fld_ptr(fldTEXT));
			break;
	   	case mnuKeyboard:
			SysKeyboardDialog(kbdDefault);
			break;
		 case mnuGraffiti:
			SysGraffitiReferenceDialog(referenceDefault);
			break;
	   	case mnuFont0:
			FldSetFont(fld_ptr(fldTEXT), (lastDocFont = stdFont));
			lastDocPageSize = 10;
			break;
	   	case mnuFont1:
			FldSetFont(fld_ptr(fldTEXT), (lastDocFont = boldFont));
			lastDocPageSize = 10;
			break;
	   	case mnuFont2:
			FldSetFont(fld_ptr(fldTEXT), (lastDocFont = largeFont));
			lastDocPageSize = 8;
			break;
	   	case mnuFont7:
			FldSetFont(fld_ptr(fldTEXT), (lastDocFont = largeBoldFont));
			lastDocPageSize = 7;
			break;
	   	case mnuFontHK:
			if	( os_charset == enc_utf8 )	{
				FldSetFont(fld_ptr(fldTEXT), (lastDocFont = HK40CustomFont));
				lastDocPageSize = 20;
				}
			else
				FrmAlert(AlertCFontMB);
			break;
	   	case mnuFontSM10:
			if	( os_charset == enc_utf8 )	{
				FldSetFont(fld_ptr(fldTEXT), (lastDocFont = SM10CustomFont));
				lastDocPageSize = 13;
				}
			else
				FrmAlert(AlertCFontMB);
			break;
	   	case mnuFontDOS:
			if	( os_charset == enc_utf8 )	{
				FldSetFont(fld_ptr(fldTEXT), (lastDocFont = DOSCustomFont));
				lastDocPageSize = 15;
				}
			else
				FrmAlert(AlertCFontMB);
			break;
	   	case mnuFontDOS4:
			if	( os_charset == enc_utf8 )	{
				FldSetFont(fld_ptr(fldTEXT), (lastDocFont = DOS4CustomFont));
				lastDocPageSize = 15;
				}
			else
				FrmAlert(AlertCFontMB);
			break;
		case	mnuDelete:
			FileDelete(0, curFSVName);

			lastFSV = 0;
			for ( i = 0; i < ufstCount; i ++ )	{
				if	( StrCompare(curFSVName, ufstTable[i]) == 0 )	{
					lastFSV = i;
					break;
					}
				}

			list = GetListPtr(lstSection);
			LstSetListChoices(list, ufstTable, ufstCount);
			UpdatePopup(trgSection, lstSection, lastFSV, ufstTable);

			StrCopy(curFSVName, ufstTable[lastFSV]);
			dispFSV(curFSVName);
			setTextPos(lastDocPos);
			break;
		  	};
    	handled = true;
		break;

	case keyDownEvent:
		switch ( e->data.keyDown.chr )	{
		case	pageUpChr: /* Prev */
			if ( FldScrollable(fld_ptr(fldTEXT), winUp) ) 
				FldScrollField(fld_ptr(fldTEXT), lastDocPageSize, winUp);
			handled = true;
			break;
		case pageDownChr: /* next */
			if	( FldScrollable(fld_ptr(fldTEXT), winDown) )	
				FldScrollField(fld_ptr(fldTEXT), lastDocPageSize, winDown);
			handled = 1;
			break;
			};
		break;
    case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
		case	btnSFT:
			FldInsert(fld_ptr(fldTEXT), "\t", 1);
	    	handled = true;
			break;
		case	btnCLOSE:
			lastDocPos = getTextPos();
			FSVSaveStatus(curFSVName, lastDocPos, lastDocFont);
			checkSaveFSV();

			closeFSV();
			FrmGotoForm(MainForm);
	    	handled = true;
			break;
		case	btnUP:
			if ( FldScrollable(fld_ptr(fldTEXT), winUp) ) 
				FldScrollField(fld_ptr(fldTEXT), 1, winUp);
	    	handled = true;
			break;
		case	btnDOWN:
			if	( FldScrollable(fld_ptr(fldTEXT), winDown) )	
				FldScrollField(fld_ptr(fldTEXT), 1, winDown);
	    	handled = true;
			break;
		case	btnPGUP:
			if ( FldScrollable(fld_ptr(fldTEXT), winUp) ) 
				FldScrollField(fld_ptr(fldTEXT), lastDocPageSize, winUp);
	    	handled = true;
			break;
		case	btnPGDN:
			if	( FldScrollable(fld_ptr(fldTEXT), winDown) )	
				FldScrollField(fld_ptr(fldTEXT), lastDocPageSize, winDown);
	    	handled = true;
			break;
		case	btnSAVE:
			lastDocPos = getTextPos();
			FSVSaveStatus(curFSVName, lastDocPos, lastDocFont);
			saveFSV(curFSVName);
			FldSetDirty(fld_ptr(fldTEXT), 0);
	    	handled = true;
			break;
			}
		break;		   

    default:
        break;
    	}

    return handled;
}

Boolean		FSViewFormHandleEvent(EventPtr e)
{
	return FSVFormHandleEventSEG(e);
}

//// OUTPUT /////////////////////////////////////////////////////////////////////////////

/*
*	CALLED FROM BRUN
*/
int		BRUNHandleEvent(EventPtr e)
{
	int		rv = 0;

    switch (e->eType) {
	case frmCloseEvent:
		if	( brun_status() == BRUN_RUNNING )	
			brun_stop();
		break;

    case menuEvent:
		switch(e->data.menu.itemID) {
		case	mnuRun:
			brun_stop();
			EvtAddEventToQueue(e);
			rv = 1;
			break;
		case	mnuBreak:
			brun_break();
			rv = 1;
			break;
		case	mnuClose:
			brun_stop();
			EvtAddEventToQueue(e);
			rv = 1;
			break;
		case mnuAbout:
			DoAboutForm();
	      	break;
		case mnuHelpFS:
			GPlmAlert(HelpFSStrTID, HelpFSStringID);
			break;
		case mnuHelpMisc:
			GPlmAlert(HelpMiscStrTID, HelpMiscStringID);
			break;
		case mnuHelpSys:
			GPlmAlert(HelpSysStrTID, HelpSysStringID);
			break;
		case mnuHelp:
			GPlmAlert(HelpStrTID, HelpStringID);
			break;
		case mnuHelpMS:
			GPlmAlert(HelpMSStrTID, HelpMSStringID);
			break;
		case mnuHelpCon:
			GPlmAlert(HelpConStrTID, HelpConStringID);
			break;
		case mnuHelpGraf:
			GPlmAlert(HelpGrafStrTID, HelpGrafStringID);
			break;
			}
		break;

	default:
		break;
    	}

    return rv;
}

//
Boolean		OutFormHandleEventSEG(EventPtr e) SEC(IDE);
Boolean		OutFormHandleEventSEG(EventPtr e)
{
	Boolean		handled = false;

    switch (e->eType) {
    case frmOpenEvent:
		dev_cls();
		brun(curFileName);
		handled = true;
		break;

	case frmCloseEvent:
		if	( brun_status() == BRUN_RUNNING )	
			brun_stop();
		handled = true;
		break;

    case menuEvent:

		switch(e->data.menu.itemID) {
		case	mnuRun:
			brun(curFileName);
			handled = true;
			break;
		case	mnuClose:
			if	( autoReturnToForm )
				FrmGotoForm(autoReturnToForm);
			else
				FrmGotoForm(MainForm);
			handled = true;
			break;
		case mnuAbout:
			DoAboutForm();
			handled = true;
			break;
		case mnuHelpFS:
			GPlmAlert(HelpFSStrTID, HelpFSStringID);
			handled = true;
			break;
		case mnuHelpMisc:
			GPlmAlert(HelpMiscStrTID, HelpMiscStringID);
			handled = true;
			break;
		case mnuHelpSys:
			GPlmAlert(HelpSysStrTID, HelpSysStringID);
			handled = true;
			break;
		case mnuHelp:
			GPlmAlert(HelpStrTID, HelpStringID);
			handled = true;
			break;
		case mnuHelpMS:
			GPlmAlert(HelpMSStrTID, HelpMSStringID);
			handled = true;
			break;
		case mnuHelpCon:
			GPlmAlert(HelpConStrTID, HelpConStringID);
			handled = true;
			break;
		case mnuHelpGraf:
			GPlmAlert(HelpGrafStrTID, HelpGrafStringID);
			handled = true;
			break;
			}
		break;

	default:
		break;
    	}

	return handled;
}

Boolean		OutFormHandleEvent(EventPtr e)
{
	return	OutFormHandleEventSEG(e);
}

//// MAIN/SELECT /////////////////////////////////////////////////////////////////////////////

static void refreshDocList() SEC(IDE);
static void refreshDocList()
{
	ListPtr	list;
	buildDocList();
	list = GetListPtr(lstBROWSE);
	LstSetListChoices(list, docTable, docCount); 
	LstDrawList(list);
}

static int doFileDetails(char *file) SEC(IDE);
static int doFileDetails(char *file)
{
	FormPtr	frmPrev, frmDet;
	ListPtr	list;
	ControlPtr	popup;
	int		hit;

	frmPrev = FrmGetActiveForm();
	frmDet = FrmInitForm(DetForm);
	FrmSetActiveForm(frmDet);

	FrmSetTitle(frmDet, file);

	list = GetListPtr(lstCat);
	LstSetListChoices(list, catList, catCount);
	LstSetSelection(list, curCat);
	popup = (ControlPtr) FrmGetObjectPtr(frmDet, FrmGetObjectIndex(frmDet, trgCat));
	CtlSetLabel(popup, catList[curCat]);

	hit = FrmDoDialog(frmDet);

	if	( hit == btnOK )	{
		// get vals
		updateDocCat(file, LstGetSelection(list));
		}

	if	( frmPrev )
		FrmSetActiveForm(frmPrev);
	FrmDeleteForm(frmDet);
	return ( hit == btnOK );
}

static int doCatEdit() SEC(IDE);
static int doCatEdit()
{
	FormPtr	frmPrev, frmDet;
	ListPtr	list;
	int		hit;

	frmPrev = FrmGetActiveForm();
	frmDet = FrmInitForm(EditCatForm);
	FrmSetActiveForm(frmDet);

	list = GetListPtr(lstCat);
	LstSetListChoices(list, catList, catCount);
	LstSetSelection(list, curCat);

	hit = FrmDoDialog(frmDet);

	if	( hit == btnOK )	{
		// get vals
		;
		}

	if	( frmPrev )
		FrmSetActiveForm(frmPrev);
	FrmDeleteForm(frmDet);
	return ( hit == btnOK );
}

static Boolean MainFormHandleEventSEG(EventPtr e) SEC(IDE);
static Boolean MainFormHandleEventSEG(EventPtr e)
{
    Boolean handled = false;
    FormPtr frm = NULL;
	FormPtr	frmpop = NULL;
	ListPtr	list;
	char	*fileName;
	static int lastFile;
	LocalID	LID;
	FormActiveStateType	stateP;
    
    switch (e->eType) {
    case frmOpenEvent:
		frm = FrmGetActiveForm();
		FrmDrawForm(frm);

		curCat = sbpref.lastCat;

		buildDocList();

		sbpref.lastDocForm = MainForm;

		if	( docCount )	{
			list = GetListPtr(lstBROWSE);
			LstSetListChoices(list, docTable, docCount); 
			LstDrawList(list);
			SetListSel(lstBROWSE, lastFile);
			}

		list = GetListPtr(lstCat);
		LstSetListChoices(list, catList, catCount);
		UpdatePopup(trgCat, lstCat, curCat, catList);

		autoReturnToForm = 0;
		handled = true;
		break;

	case frmCloseEvent:
		destroyDocList();
		break;

	case	evtRefreshDOCs:
		refreshDocList();
		lastFile = 0;
		list = GetListPtr(lstBROWSE);
		LstSetListChoices(list, docTable, docCount); 
		LstDrawList(list);
		SetListSel(lstBROWSE, lastFile);
		break;

    case menuEvent:
		switch(e->data.menu.itemID) {
		case mnuSBPREF:
			frm = FrmGetActiveForm();
			FrmSaveActiveState(&stateP);

			frmpop = FrmInitForm(SBPREF_FORM);
			FrmSetActiveForm(frmpop);
			{
				ControlPtr	dellog;
				ListPtr		list;
				ControlPtr	popup;
				char		num[33];

				dellog = FrmGetObjectPtr(frmpop, FrmGetObjectIndex(frmpop, chkDELLOG));
				CtlSetValue(dellog, ((sbpref.flags & 1) != 0));

				list = (ListPtr) FrmGetObjectPtr(frmpop, FrmGetObjectIndex(frmpop, lstCHAR));
//				LstMakeItemVisible(list, sbpref.charset);

				popup = (ControlPtr) FrmGetObjectPtr(frmpop, FrmGetObjectIndex(frmpop, trgCHAR));
				CtlSetLabel(popup, LstGetSelectionText(list, sbpref.charset));
//				CtlDrawControl(popup);

				StrPrintF(num, "%d", sbpref.cclabs1);
				fld_setText(fldCLABS1, num, 0);
				StrPrintF(num, "%d", sbpref.ccpass2);
				fld_setText(fldCPASS2, num, 0);

				if	( FrmDoDialog(frmpop) == btnOK )	{
					sbpref.charset = LstGetSelection(list);
					sbpref.flags = 0;

					if	( CtlGetValue(dellog) )	
						sbpref.flags |= 1;

					sbpref.cclabs1 = (word) xstrtol(fld_getTextPtr(fldCLABS1));
					sbpref.ccpass2 = (word) xstrtol(fld_getTextPtr(fldCPASS2));
					if	( sbpref.cclabs1 < 64 )		sbpref.cclabs1 = 64;
					if	( sbpref.ccpass2 < 64 )		sbpref.ccpass2 = 64;

					savePrefs();
					}
			}

			FrmRestoreActiveState(&stateP);
			FrmSetActiveForm(frm);
			FrmDeleteForm(frmpop);

			os_charset = sbpref.charset;
			os_cclabs1 = sbpref.cclabs1;
			os_ccpass2 = sbpref.ccpass2;

//			FrmUpdateForm(MainForm, frmRedrawUpdateCode);
			break;
		case mnuClose:
			Exec("lnch");
			break;
		case mnuAbout:
			DoAboutForm();
	      	break;
		case mnuHelp:
			GPlmAlert(HelpStrTID, HelpStringID);
			break;
		case mnuHelpFS:
			GPlmAlert(HelpFSStrTID, HelpFSStringID);
			break;
		case mnuHelpMisc:
			GPlmAlert(HelpMiscStrTID, HelpMiscStringID);
			break;
		case mnuHelpSys:
			GPlmAlert(HelpSysStrTID, HelpSysStringID);
			break;
		case mnuHelpMS:
			GPlmAlert(HelpMSStrTID, HelpMSStringID);
			break;
		case mnuHelpCon:
			GPlmAlert(HelpConStrTID, HelpConStringID);
			break;
		case mnuHelpGraf:
			GPlmAlert(HelpGrafStrTID, HelpGrafStringID);
			break;
		case	mnuBeam:
			if	( docCount )	{
				fileName = docTable[lastFile];
				beamSendDB(fileName);
				}
			break;
		case	mnuDetails:
			if	( docCount )	{
				fileName = docTable[lastFile];
				if	( doFileDetails(fileName) && curCat != 0 )	{
					refreshDocList();

					lastFile = 0;
					list = GetListPtr(lstBROWSE);
					LstSetListChoices(list, docTable, docCount); 
					LstDrawList(list);
					SetListSel(lstBROWSE, lastFile);
					}
				}
			break;
		case	mnuFSV:
			FrmGotoForm(FSVForm);
			break;
		case	mnuNew:
			if	( BasicInputBox("New FileName:", temp, 255) )	{
				fileName = temp;
				mkPutFileExt(temp);
				createNewDOC(fileName);
				if	( dbExist(fileName) )	{
					StrCopy(curFileName, fileName);
					FrmGotoForm(ViewForm);
					}
				}
			break;
		case	mnuEditCat:
			doCatEdit();
			break;
		case	mnuEdit:
			if	( docCount )	{
				fileName = docTable[lastFile];
				if	( dbExist(fileName) )	{
					StrCopy(curFileName, fileName);
					FrmGotoForm(ViewForm);
					}
				}
			break;
		case	mnuCopyFile:
			StrCopy(curFileName, docTable[lastFile]);
			if	( BasicInputBox("New FileName:", temp, 255) )	{
				mkPutFileExt(temp);
				fileName = temp;
				copyDOC(curFileName, fileName);
				refreshDocList();
				}
			break;
		case	mnuRename:
			StrCopy(curFileName, docTable[lastFile]);
			if	( BasicInputBox("New FileName:", temp, 255) )	{
				mkPutFileExt(temp);
				fileName = temp;

				renameDOC(curFileName, fileName);
				refreshDocList();

				lastFile = 0;
				list = GetListPtr(lstBROWSE);
				LstSetListChoices(list, docTable, docCount); 
				LstDrawList(list);
				SetListSel(lstBROWSE, lastFile);
				}
			break;
		case	mnuRun:
			autoReturnToForm = MainForm;
			if	( docCount )	{
				fileName = docTable[lastFile];
				StrCopy(curFileName, fileName);
				FrmGotoForm(SBOutForm);
				}
			break;

		case	mnuDelete:
			if	( docCount )	{
				fileName = docTable[lastFile];
				if	( FrmCustomAlert(ConfAlertID, "Delete '", fileName, "' ?") == 0 )	{
					fileName = docTable[lastFile];
					LID = DmFindDatabase(0, fileName);
					if	( LID )	{
						DmDeleteDatabase(0, LID);
						refreshDocList();

						lastFile = 0;
						list = GetListPtr(lstBROWSE);
						LstSetListChoices(list, docTable, docCount); 
						LstDrawList(list);
						SetListSel(lstBROWSE, lastFile);
						}
					else
						FrmCustomAlert(InfoAlertID, fileName, ": The PalmOS API routine DmFindDatabase() FAILED!!!", "");
					}
				}
			break;
			}

    	handled = true;
		break;

	case popSelectEvent:
    	switch (e->data.popSelect.controlID) {
	    case trgCat:
			if	( curCat != e->data.popSelect.selection )	{
				curCat = e->data.popSelect.selection;
				sbpref.lastCat = curCat;

				refreshDocList();
				lastFile = 0;
				list = GetListPtr(lstBROWSE);
				LstSetListChoices(list, docTable, docCount); 
				LstDrawList(list);
				SetListSel(lstBROWSE, lastFile);

				UpdatePopup(trgCat, lstCat, curCat, catList);
				}
			handled = 1;
			break;
			}
		break;

    case lstSelectEvent:
		switch(e->data.lstSelect.listID) {
		case lstBROWSE:
			lastFile = GetListSel(lstBROWSE);
    		handled = true;
			break;
			}
		break;

    case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
		case	btnCLOSE:
			Exec("lnch");
	    	handled = true;
			break;
		case	btnOPEN:
			if	( docCount )	{
				fileName = docTable[lastFile];
				if	( dbExist(fileName) )	{
					StrCopy(curFileName, fileName);
					FrmGotoForm(ViewForm);
					}
				}
	    	handled = true;
			break;
		case	btnNEW:
			if	( BasicInputBox("New FileName:", temp, 255) )	{
				fileName = temp;
				mkPutFileExt(temp);
				createNewDOC(fileName);
				if	( dbExist(fileName) )	{
					StrCopy(curFileName, fileName);
					FrmGotoForm(ViewForm);
					}
				}
	    	handled = true;
			break;
		case	btnRUN:
			autoReturnToForm = MainForm;
			if	( docCount )	{
				fileName = docTable[lastFile];
				StrCopy(curFileName, fileName);
				FrmGotoForm(SBOutForm);
				}
	    	handled = true;
			break;

		case	btnDELETE:
			if	( docCount )	{
				fileName = docTable[lastFile];
				if	( FrmCustomAlert(ConfAlertID, "Delete '", fileName, "' ?") == 0 )	{
					fileName = docTable[lastFile];
					LID = DmFindDatabase(0, fileName);
					if	( LID )	{
						DmDeleteDatabase(0, LID);
						refreshDocList();

						lastFile = 0;
						list = GetListPtr(lstBROWSE);
						LstSetListChoices(list, docTable, docCount); 
						LstDrawList(list);
						SetListSel(lstBROWSE, lastFile);
						}
					else
						FrmCustomAlert(InfoAlertID, fileName, ": The PalmOS API routine DmFindDatabase() FAILED!!!", "");
					}
				}
			
	    	handled = true;
			break;
			}
		break;

    default:
        break;
    	}

    return handled;
}

Boolean MainFormHandleEvent(EventPtr e)
{
	return MainFormHandleEventSEG(e);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;

    if (e->eType == frmLoadEvent) {
		formId = e->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);

		switch(formId) {
		case MainForm:
		    FrmSetEventHandler(frm, MainFormHandleEvent);
		    break;
		case ViewForm:
		    FrmSetEventHandler(frm, ViewFormHandleEvent);
		    break;
		case FSVForm:
		    FrmSetEventHandler(frm, FSViewFormHandleEvent);
		    break;
		case SBOutForm:
		    FrmSetEventHandler(frm, OutFormHandleEvent);
		    break;
			}
		return true;
	    }

    return false;
}

void	freeCatList() SEC(IDE);
void	freeCatList()
{
	int		i;

	if	( catList )	{
		for ( i = 0; i < catCount; i ++ )	
			tmp_free(catList[i]);

		tmp_free(catList);

		catList = NULL;
		catCount = 0;
		}
}

void	loadCategories() SEC(IDE);
void	loadCategories()
{
	file_t	f;
	int		i, recs;
	byte	*buff;

	freeCatList();

	buff = tmp_alloc(64);
	if	( !dbExist("SmBa.IDE.CAT") )	{
		f = dbOpenAlw("SmBa.IDE.CAT", SmBa, idDATA);

		StrCopy(buff, "All");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		StrCopy(buff, "Business");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		StrCopy(buff, "Personal");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		StrCopy(buff, "Library");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		StrCopy(buff, "Graphics");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		StrCopy(buff, "Math.");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		StrCopy(buff, "System");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		StrCopy(buff, "Temp");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		StrCopy(buff, "Unfiled");
		dbWrite(f, 0xFFFF, buff, StrLen(buff)+1);

		dbClose(f);
		}

	f = dbOpen("SmBa.IDE.CAT");

	recs = DmNumRecords(f);
	catList = tmp_alloc(sizeof(char_p) * recs);
	catCount = 0;
	for ( i = 0; i < recs; i ++ )	{
		dbRead(f, i, buff, 64);

		if	( StrCompare(buff, "Unfiled") == 0 )	
			catUnfiledID = i;

		catList[catCount] = tmp_alloc(StrLen(buff)+1);
		StrCopy(catList[catCount], buff);
		catCount ++;
		}

	tmp_free(buff);
	dbClose(f);
}

/*
*/
void loadPrefs() SEC(IDE);
void loadPrefs()
{
	word	prefSize, prefVer;

	prefSize = sizeof(sbpref_t);
	prefVer = PrefGetAppPreferences(SmBa, 0, &sbpref, &prefSize, 1);
	if	( prefVer == noPreferenceFound )	{
		sbpref.lastDocSec = 0;
		sbpref.lastDocPos = 0;
		sbpref.lastDocForm = 0;
		sbpref.font = stdFont;
		sbpref.lastCat = curCat;
		sbpref.charset = 0;
		sbpref.flags = 0x1;
		sbpref.cclabs1 = sbpref.ccpass2 = 256;
		}

	if	( sbpref.cclabs1 <= 0 )
		sbpref.cclabs1 = 256;
	if	( sbpref.ccpass2 <= 0 )
		sbpref.ccpass2 = 256;

	os_charset = sbpref.charset;
	os_cclabs1 = sbpref.cclabs1;
	os_ccpass2 = sbpref.ccpass2;
}

/*
*/
void savePrefs() SEC(IDE);
void savePrefs()
{
	StrCopy(sbpref.lastFileName, curFileName);
	sbpref.lastCat = curCat;
	sbpref.lastDocSec = lastDocSec;
	sbpref.lastDocPos = lastDocPos;
//	sbpref.lastDocForm = ViewForm;
	sbpref.font = lastDocFont;
	PrefSetAppPreferences(SmBa, 0, 1, &sbpref, sizeof(sbpref_t), 1);
}

/*
*	setup "color" screen mode
*/
void	InitialScreenMode() SEC(IDE);
void	InitialScreenMode()
{
	DWord	NewDepth;
	DWord	SupportDepth = 1;
	Boolean	SupportColor_flag = 0;

	FtrGet(sysFtrCreator, sysFtrNumDisplayDepth, &OldScreenDepth);
	ScrDisplayMode(scrDisplayModeGetSupportedDepths, NULL, NULL, &SupportDepth,NULL);
	ScrDisplayMode(scrDisplayModeGetSupportsColor, NULL, NULL, NULL, &SupportColor_flag);

	if      (SupportDepth & 0x0080) NewDepth = 8;
	else if (SupportDepth & 0x0008) NewDepth = 4;
	else if (SupportDepth & 0x0002) NewDepth = 2;
	else                            NewDepth = 1;

	if ( !FtrSet(sysFtrCreator, sysFtrNumDisplayDepth, NewDepth) ) 	
		ScrDisplayMode(scrDisplayModeSetToDefaults,NULL,NULL,NULL,NULL);
	else	{
		ScrDisplayMode(scrDisplayModeSet,NULL,NULL,&NewDepth,NULL);
		WinScreenMode(winScreenModeSet, 0, 0, &NewDepth, 0);
		}
}

/*
*	restore screen mode
*/
void	RecoverScreenMode() SEC(IDE);
void	RecoverScreenMode()
{
	FtrSet(sysFtrCreator, sysFtrNumDisplayDepth, OldScreenDepth);
	ScrDisplayMode(scrDisplayModeSetToDefaults, NULL, NULL, NULL, NULL);
}

/*
*	load buildin fonts
*/
void	loadFonts() SEC(IDE);
void	loadFonts()
{
	gHK40FontHandle = DmGetResource('pFNT', HK40FONT);
	FntDefineFont(HK40CustomFont, MemHandleLock(gHK40FontHandle));
	gSM10FontHandle = DmGetResource('NFNT', SM10FONT);
	FntDefineFont(SM10CustomFont, MemHandleLock(gSM10FontHandle));
	gDOSFontHandle = DmGetResource('NFNT', DOSFONT);
	FntDefineFont(DOSCustomFont, MemHandleLock(gDOSFontHandle));
	gDOS4FontHandle = DmGetResource('NFNT', DOS4FONT);
	FntDefineFont(DOS4CustomFont, MemHandleLock(gDOS4FontHandle));
}

/*
*	load buildin fonts
*/
void	unloadFonts() SEC(IDE);
void	unloadFonts()
{
	MemHandleUnlock(gHK40FontHandle);		DmReleaseResource(gHK40FontHandle);
	MemHandleUnlock(gSM10FontHandle);		DmReleaseResource(gSM10FontHandle);
	MemHandleUnlock(gDOSFontHandle);		DmReleaseResource(gDOSFontHandle);
	MemHandleUnlock(gDOS4FontHandle);		DmReleaseResource(gDOS4FontHandle);
}

/*
*/
Word StartApplication(void) SEC(IDE);
Word StartApplication(void)
{
	Err	error;

	InitialScreenMode();
	loadCategories();
	error = SysLibFind(MathLibName, &MathLibRef);
	if	( error )	
		SysLibLoad(LibType, MathLibCreator, &MathLibRef);

	error = MathLibOpen(MathLibRef, MathLibVersion);
	if	( error )	{
		FrmAlert(AlertMathLib);
		RecoverScreenMode();
		return 0;
		}
	
	loadFonts();
	loadPrefs();

	if	( sbpref.flags & 1 )
		FileDelete(0, "SB.LOG");

	if	( sbpref.lastDocForm == ViewForm && dbExist(sbpref.lastFileName) )	{
		StrCopy(curFileName, sbpref.lastFileName);
		FrmGotoForm(ViewForm);
		}
	else
	    FrmGotoForm(MainForm);
    return 1;
}

/*
*/
void StopApplication(void) SEC(IDE);
void StopApplication(void)
{
	Int		librefs;

	closeDOC();
	savePrefs();
    FrmSaveAllForms();
    FrmCloseAllForms();

	freeCatList();
	
	unloadFonts();

	MathLibClose(MathLibRef, &librefs);
	if	( librefs == 0 )
		SysLibRemove(MathLibRef);

	RecoverScreenMode();
}

/*
*/
void EventLoop(void)
{
    Word err;
    EventType e;

    do {
		EvtGetEvent(&e, evtWaitForever);
		if (! SysHandleEvent (&e))
		    if (! MenuHandleEvent (NULL, &e, &err))
			if (! ApplicationHandleEvent (&e))
			    FrmDispatchEvent (&e);
	    } while (e.eType != appStopEvent);
}

/*
*/
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word	err;

	if ( cmd == sysAppLaunchCmdOpenDB ) {
		/*
		*	Command-line args in PalmOS! :)
		*/
		word 	card = ((SysAppLaunchCmdOpenDBType *) cmdPBP)->cardNo;
		LocalID lid  = ((SysAppLaunchCmdOpenDBType *) cmdPBP)->dbID;
		dword	type, creator;

        char startupDocName[dmDBNameLength];
        DmDatabaseInfo(card, lid, startupDocName, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &type, &creator);
        if ( !(type == 'TEXT' && creator == 'SmBa') ) {
			FrmAlert(AlertArgOpen);
			return 1;
	        }

		// update prefs - auto load the new file
		loadPrefs();
		sbpref.lastDocSec = 0;
		sbpref.lastDocPos = 0;
		sbpref.font = stdFont;
		sbpref.lastDocForm = ViewForm;
		strcpy(sbpref.lastFileName, startupDocName);
		PrefSetAppPreferences(SmBa, 0, 1, &sbpref, sizeof(sbpref_t), 1);

		// normal launch
        cmd = sysAppLaunchCmdNormalLaunch;
		}

    if ( cmd == sysAppLaunchCmdNormalLaunch )	{
		/*
		*	Normal launch
		*/
		ExgRegisterData(SmBa, exgRegExtensionID, "bas");		//	Register .bas files
		if	( StartApplication() )	{
			EventLoop();
			StopApplication();
			}
	    } 
	else if (cmd == sysAppLaunchCmdSyncNotify) {
		ExgRegisterData(SmBa, exgRegExtensionID, "bas");		//	Register .bas files
		}
	else if (cmd == sysAppLaunchCmdExgAskUser)    {
		ExgAskParamPtr paramP = (ExgAskParamPtr)cmdPBP;
       	paramP->result = exgAskOk; // paramP->result = exgAskCancel;
        }
	else if (cmd == sysAppLaunchCmdExgReceiveData)    {
		byte prog_running = ((launchFlags & sysAppLaunchFlagSubCall) != 0);

		if	( (err = beamGetDB((ExgSocketPtr)cmdPBP, prog_running)) != 0 )
			return err;
		}

    return 0;
}

